"""
Automated search for capacity for different systems via latency vs qps data.
A system is characterised by:
1. trace
2. model
3. sku
4. scheduler
"""

import argparse
import json
import os
import time

import yaml

from vidur.config_optimizer.config_explorer.config_explorer import ConfigExplorer
from vidur.logger import init_logger

logger = init_logger(__name__)


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--num-threads", type=int, default=None)
    parser.add_argument(
        "--min-search-granularity",
        type=float,
        default=2.5,
        help="Minimum search granularity for capacity (%)",
    )
    parser.add_argument("--output-dir", type=str, required=True)
    parser.add_argument("--cache-dir", type=str, default="./cache_tmpfs")
    parser.add_argument("--config-path", type=str, required=True)
    parser.add_argument("--scheduling-delay-slo-value", type=float, default=5.0)
    parser.add_argument("--scheduling-delay-slo-quantile", type=float, default=0.99)
    parser.add_argument("--max-iterations", type=int, default=20)
    parser.add_argument(
        "--time-limit", type=int, default=30, help="Time limit in minutes"
    )
    parser.add_argument("--debug", action="store_true")
    parser.add_argument("--skip-cache-warmup", action="store_true")

    args = parser.parse_args()

    default_num_threads = os.cpu_count() - 2
    if args.num_threads is not None:
        args.num_threads = min(args.num_threads, default_num_threads)
    else:
        args.num_threads = default_num_threads

    return args


if __name__ == "__main__":
    args = get_args()

    config = yaml.safe_load(open(args.config_path))

    assert (
        args.scheduling_delay_slo_quantile >= 0
        and args.scheduling_delay_slo_quantile <= 1
    )

    os.makedirs(args.output_dir, exist_ok=True)

    logger.info("Starting config optimizer")
    logger.info(f"Args: {args}")
    logger.info(f"Config: {config}")

    # store the config and args
    json.dump(vars(args), open(f"{args.output_dir}/args.json", "w"))
    json.dump(config, open(f"{args.output_dir}/config.json", "w"))

    multiple_capacity_search = ConfigExplorer(args, config)

    start_time = time.time()

    all_results = multiple_capacity_search.run()

    end_time = time.time()

    logger.info(f"Simulation took time: {end_time - start_time}")



import json
from abc import ABC, abstractmethod
from typing import List

from vidur.config import BaseRequestGeneratorConfig
from vidur.entities import Request


class BaseRequestGenerator(ABC):

    def __init__(self, config: BaseRequestGeneratorConfig):
        self.config = config

    @abstractmethod
    def generate_requests(self) -> List[Request]:
        pass

    def generate(self) -> List[Request]:
        requests = self.generate_requests()
        return requests



import math
import random

from vidur.config import PoissonRequestIntervalGeneratorConfig
from vidur.request_generator.base_request_interval_generator import (
    BaseRequestIntervalGenerator,
)


class PoissonRequestIntervalGenerator(BaseRequestIntervalGenerator):

    def __init__(self, config: PoissonRequestIntervalGeneratorConfig):
        super().__init__(config)

        self.qps = self.config.qps
        self.std = 1.0 / self.qps
        self.max_interval = self.std * 3.0

    def get_next_inter_request_time(self) -> float:
        next_interval = -math.log(1.0 - random.random()) / self.qps
        next_interval = min(next_interval, self.max_interval)

        return next_interval



import argparse
import copy

import ray

from vidur.config_optimizer.config_explorer.capacity_search import CapacitySearch
from vidur.config_optimizer.config_explorer.config import JobConfig
from vidur.config_optimizer.config_explorer.ray_utils import (
    CpuAssignmentManager,
    RayParallelRunner,
    run_on_each_node,
)


def run_search(
    job_config: JobConfig,
    args: argparse.Namespace,
    cpu_core_assignment_manager: CpuAssignmentManager = None,
    cpu_core_id: int = None,
):
    capacity_search = CapacitySearch(
        job_config,
        args,
        cpu_core_assignment_manager,
        cpu_core_id,
    )
    return capacity_search.search()


class ConfigExplorer:
    def __init__(
        self,
        args: argparse.Namespace,
        config: dict,
    ):
        self.args = args
        self.config = config

        ray.init(ignore_reinit_error=True)

    def _warmup_cache(self):
        job_configs = JobConfig.generate_unique_model_job_configs(self.config)

        args_for_warmup = copy.deepcopy(self.args)
        args_for_warmup.max_iterations = 1

        for job_config in job_configs:
            all_node_results = run_on_each_node(
                run_search,
                job_config,
                args_for_warmup,
            )
            assert all(all_node_results) or not any(
                all_node_results
            ), "All nodes should have the same result"

    def run(self):
        if not self.args.skip_cache_warmup:
            self._warmup_cache()

        job_configs = JobConfig.generate_job_configs(self.config)

        ray_parallel_runner = RayParallelRunner()

        remote_func = (
            lambda cpu_core_assignment_manager, cpu_core_id, job_config: run_search(
                job_config,
                self.args,
                cpu_core_assignment_manager,
                cpu_core_id,
            )
        )
        all_results = ray_parallel_runner.map(
            remote_func,
            job_configs,
        )
        return all_results



from typing import Any

from vidur.entities import Request
from vidur.scheduler.replica_scheduler.mnemosyne_base_replica_scheduler import (
    BatchFormationTracker,
    MnemosyneBaseReplicaScheduler,
)


class MnemosyneFCFSFixedChunkReplicaScheduler(MnemosyneBaseReplicaScheduler):
    def _get_request_next_num_q_tokens(
        self,
        time: float,
        request: Request,
        batch_formation_tracker: BatchFormationTracker,
    ) -> int:
        assert not request.completed

        if self._is_prefill_complete(request):
            return 1

        active_kvp_groups = self._get_active_kvp_group_ids(request)
        batched_num_q_tokens_across_groups = (
            batch_formation_tracker.get_q_tokens_for_kvp_groups(active_kvp_groups)
        )
        max_num_q_tokens_across_groups = max(batched_num_q_tokens_across_groups)

        next_num_tokens = min(
            request.num_prefill_tokens - self._get_num_processed_tokens(request),
            self._config.chunk_size - max_num_q_tokens_across_groups,
        )

        next_num_tokens = max(0, next_num_tokens)

        return next_num_tokens

    def _get_request_priority(self, request: Request) -> Any:
        return (request.arrived_at, request.id)



import json
from argparse import (
    ArgumentDefaultsHelpFormatter,
    ArgumentParser,
    BooleanOptionalAction,
)
from collections import defaultdict, deque
from dataclasses import MISSING, fields, make_dataclass
from typing import Any, get_args

from vidur.config.base_poly_config import BasePolyConfig
from vidur.config.utils import (
    get_all_subclasses,
    get_inner_type,
    is_bool,
    is_composed_of_primitives,
    is_dict,
    is_list,
    is_optional,
    is_primitive_type,
    is_subclass,
    to_snake_case,
)


def topological_sort(dataclass_dependencies: dict) -> list:
    in_degree = defaultdict(int)
    for cls, dependencies in dataclass_dependencies.items():
        for dep in dependencies:
            in_degree[dep] += 1

    zero_in_degree_classes = deque(
        [cls for cls in dataclass_dependencies if in_degree[cls] == 0]
    )
    sorted_classes = []

    while zero_in_degree_classes:
        cls = zero_in_degree_classes.popleft()
        sorted_classes.append(cls)
        for dep in dataclass_dependencies[cls]:
            in_degree[dep] -= 1
            if in_degree[dep] == 0:
                zero_in_degree_classes.append(dep)

    return sorted_classes


def reconstruct_original_dataclass(self) -> Any:
    """
    This function is dynamically mapped to FlatClass as an instance method.
    """
    sorted_classes = topological_sort(self.dataclass_dependencies)
    instances = {}

    for _cls in reversed(sorted_classes):
        args = {}

        for prefixed_field_name, original_field_name, field_type in self.dataclass_args[
            _cls
        ]:
            if is_subclass(field_type, BasePolyConfig):
                config_type = getattr(self, f"{original_field_name}_type", None)
                # find all subclasses of field_type and check which one matches the config_type
                for subclass in get_all_subclasses(field_type):
                    if str(subclass.get_type()) == config_type.lower():
                        args[original_field_name] = instances[subclass]
                        break

            elif hasattr(field_type, "__dataclass_fields__"):
                args[original_field_name] = instances[field_type]
            else:
                value = getattr(self, prefixed_field_name)
                if callable(value):
                    # to handle default factory values
                    value = value()
                args[original_field_name] = value

        instances[_cls] = _cls(**args)

    return instances[sorted_classes[0]]


@classmethod
def create_from_cli_args(cls) -> Any:
    """
    This function is dynamically mapped to FlatClass as a class method.
    """
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)

    for field in fields(cls):
        nargs = None
        action = None
        field_type = field.type
        help_text = cls.metadata_mapping[field.name].get("help", None)

        if is_list(field.type):
            assert is_composed_of_primitives(field.type)
            field_type = get_args(field.type)[0]
            if is_primitive_type(field_type):
                nargs = "+"
            else:
                field_type = json.loads
        elif is_dict(field.type):
            assert is_composed_of_primitives(field.type)
            field_type = json.loads
        elif is_bool(field.type):
            action = BooleanOptionalAction

        arg_params = {
            "type": field_type,
            "action": action,
            "help": help_text,
        }

        # handle cases with default and default factory args
        if field.default is not MISSING:
            value = field.default
            if callable(value):
                value = value()
            arg_params["default"] = value
        elif field.default_factory is not MISSING:
            arg_params["default"] = field.default_factory()
        else:
            arg_params["required"] = True

        if nargs:
            arg_params["nargs"] = nargs
        parser.add_argument(f"--{field.name}", **arg_params)

    args = parser.parse_args()

    return cls(**vars(args))


def create_flat_dataclass(input_dataclass: Any) -> Any:
    """
    Creates a new FlatClass type by recursively flattening the input dataclass.
    This allows for easy parsing of command line arguments along with storing/loading the configuration to/from a file.
    """
    meta_fields_with_defaults = []
    meta_fields_without_defaults = []
    processed_classes = set()
    dataclass_args = defaultdict(list)
    dataclass_dependencies = defaultdict(set)
    metadata_mapping = {}

    def process_dataclass(_input_dataclass, prefix=""):
        if _input_dataclass in processed_classes:
            return

        processed_classes.add(_input_dataclass)

        for field in fields(_input_dataclass):
            prefixed_name = f"{prefix}{field.name}"

            if is_optional(field.type):
                field_type = get_inner_type(field.type)
            else:
                field_type = field.type

            # # if field is a BasePolyConfig, add a type argument and process it as a dataclass
            if is_subclass(field_type, BasePolyConfig):
                dataclass_args[_input_dataclass].append(
                    (field.name, field.name, field_type)
                )

                type_field_name = f"{field.name}_type"
                default_value = str(field.default_factory().get_type())
                meta_fields_with_defaults.append(
                    (type_field_name, type(default_value), default_value)
                )
                metadata_mapping[type_field_name] = field.metadata

                assert hasattr(field_type, "__dataclass_fields__")
                for subclass in get_all_subclasses(field_type):
                    dataclass_dependencies[_input_dataclass].add(subclass)
                    process_dataclass(subclass, f"{to_snake_case(subclass.__name__)}_")
                continue

            # if field is a dataclass, recursively process it
            if hasattr(field_type, "__dataclass_fields__"):
                dataclass_dependencies[_input_dataclass].add(field_type)
                dataclass_args[_input_dataclass].append(
                    (field.name, field.name, field_type)
                )
                process_dataclass(field_type, f"{to_snake_case(field_type.__name__)}_")
                continue

            field_default = field.default if field.default is not MISSING else MISSING
            field_default_factory = (
                field.default_factory
                if field.default_factory is not MISSING
                else MISSING
            )

            if field_default is not MISSING:
                meta_fields_with_defaults.append(
                    (prefixed_name, field_type, field_default)
                )
            elif field_default_factory is not MISSING:
                meta_fields_with_defaults.append(
                    (prefixed_name, field_type, field_default_factory)
                )
            else:
                meta_fields_without_defaults.append((prefixed_name, field_type))

            dataclass_args[_input_dataclass].append(
                (prefixed_name, field.name, field_type)
            )
            metadata_mapping[prefixed_name] = field.metadata

    process_dataclass(input_dataclass)

    meta_fields = meta_fields_without_defaults + meta_fields_with_defaults
    FlatClass = make_dataclass("FlatClass", meta_fields)

    # Metadata fields
    FlatClass.dataclass_args = dataclass_args
    FlatClass.dataclass_dependencies = dataclass_dependencies
    FlatClass.metadata_mapping = metadata_mapping

    # Helper methods
    FlatClass.reconstruct_original_dataclass = reconstruct_original_dataclass
    FlatClass.create_from_cli_args = create_from_cli_args

    return FlatClass



from typing import List


class SequenceProxy:
    def __init__(self, total_len: int, processed_len: int):
        self._total_len = total_len
        self._processed_len = processed_len
        return

    def get_len(self):
        return self._total_len

    def get_next_prompt_chunk_len(self, chunk_size: int):
        return min(chunk_size, self._total_len - self._processed_len)

    def get_num_prompt_tokens_processed(self):
        return self._processed_len

    def get_num_prompt_tokens_stage_processed(self):
        return self._processed_len


class SequenceMetadataProxy:
    def __init__(
        self,
        is_prompt: bool,
        total_len: int,
        processed_len: int,
        block_table: List[int],
    ):
        self.schedule_id = 0
        self.is_prompt = is_prompt
        self.prompt_chunk_len = (total_len - processed_len) if is_prompt else None
        self.seq = SequenceProxy(total_len, processed_len)
        self.block_table = block_table



import heapq
from abc import abstractmethod
from collections import OrderedDict, defaultdict
from math import ceil
from typing import Any, Dict, List, Optional, Union

from vidur.entities import (
    Batch,
    BatchStage,
    KVParallelBatch,
    KVParallelBatchStage,
    Request,
)
from vidur.logger import init_logger
from vidur.scheduler.replica_scheduler.base_replica_scheduler import (
    BaseReplicaScheduler,
)

logger = init_logger(__name__)

MAX_NUM_SKIPPED_REQUESTS = 10


class BatchFormationTracker:
    def __init__(
        self,
        replica_id: int,
        max_micro_batch_size: int,
        kv_parallel_size: int,
        max_num_tokens_per_kvp_group: int,
    ):
        self.replica_id: int = replica_id
        self.max_micro_batch_size: int = max_micro_batch_size
        self.kv_parallel_size: int = kv_parallel_size
        self.max_num_tokens_per_kvp_group: int = max_num_tokens_per_kvp_group

        self.num_requests: int = 0
        self.skipped_requests: List[Request] = []

        self.batch_requests: List[List[Request]] = [
            [] for _ in range(self.kv_parallel_size)
        ]
        self.batch_request_num_processed_tokens: List[List[int]] = [
            [] for _ in range(self.kv_parallel_size)
        ]
        self.batch_num_q_tokens: List[List[int]] = [
            [] for _ in range(self.kv_parallel_size)
        ]
        self.batch_num_kv_tokens: List[List[int]] = [
            [] for _ in range(self.kv_parallel_size)
        ]
        self.batch_num_active_kvp_groups: List[List[int]] = [
            [] for _ in range(self.kv_parallel_size)
        ]
        self.batch_last_kvp_group_ids: List[List[int]] = [
            [] for _ in range(self.kv_parallel_size)
        ]
        self.batch_total_num_q_tokens: List[int] = [
            0 for _ in range(self.kv_parallel_size)
        ]

    def _get_num_kv_tokens(
        self,
        num_processed_tokens: int,
        active_kvp_groups: List[int],
        is_last_group: bool,
    ) -> int:
        if is_last_group:
            num_kv_tokens_in_other_groups = (
                len(active_kvp_groups) - 1
            ) * self.max_num_tokens_per_kvp_group
            num_kv_tokens = num_processed_tokens - num_kv_tokens_in_other_groups
        else:
            num_kv_tokens = self.max_num_tokens_per_kvp_group

        assert num_kv_tokens >= 0

        return num_kv_tokens

    def add_request(
        self,
        request: Request,
        num_q_tokens: int,
        num_processed_tokens: int,
        active_kvp_groups: List[int],
    ) -> None:
        self.num_requests += 1

        for i, kvp_group_id in enumerate(active_kvp_groups):
            is_last_group = i == len(active_kvp_groups) - 1
            num_kv_tokens = self._get_num_kv_tokens(
                num_processed_tokens, active_kvp_groups, is_last_group
            )

            self.batch_requests[kvp_group_id].append(request)
            self.batch_num_q_tokens[kvp_group_id].append(num_q_tokens)
            self.batch_num_kv_tokens[kvp_group_id].append(num_kv_tokens)
            self.batch_num_active_kvp_groups[kvp_group_id].append(
                len(active_kvp_groups)
            )
            self.batch_request_num_processed_tokens[kvp_group_id].append(
                num_processed_tokens
            )
            self.batch_last_kvp_group_ids[kvp_group_id].append(active_kvp_groups[-1])
            self.batch_total_num_q_tokens[kvp_group_id] += num_q_tokens

    def can_add_request(self) -> bool:
        return self.num_requests < self.max_micro_batch_size

    def get_q_tokens_for_kvp_groups(self, active_kvp_groups: List[int]) -> List[int]:
        return [
            self.batch_total_num_q_tokens[kvp_group_id]
            for kvp_group_id in active_kvp_groups
        ]

    def get_batch(self) -> Optional[KVParallelBatch]:
        if self.num_requests == 0:
            return

        batch_mapping: Dict[int, Batch] = {}

        for kvp_group_id in range(self.kv_parallel_size):
            if not self.batch_num_q_tokens[kvp_group_id]:
                continue

            batch_mapping[kvp_group_id] = Batch(
                self.replica_id,
                self.batch_requests[kvp_group_id],
                self.batch_num_q_tokens[kvp_group_id],
                self.batch_num_kv_tokens[kvp_group_id],
                self.batch_num_active_kvp_groups[kvp_group_id],
                self.batch_last_kvp_group_ids[kvp_group_id],
                kvp_group_id,
            )

        return KVParallelBatch(self.replica_id, batch_mapping)


class MnemosyneBaseReplicaScheduler(BaseReplicaScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # For vLLM and its derivatives, we only need to set a loose max batch size
        # Memory requirements are handled explicitly by the scheduler
        self._max_micro_batch_size = self._config.batch_size_cap // self._num_stages
        self._watermark_blocks = int(
            self._config.watermark_blocks_fraction * self._config.num_blocks
        )

        self._kv_parallel_size = self._replica_config.kv_parallel_size

        if self._kv_parallel_size > 1:
            assert (
                self._replica_config.max_num_tokens_per_kvp_group
                > self._replica_config.block_size
            ), "max_num_tokens_per_kvp_group should be greater than block_size"
            assert (
                self._replica_config.max_num_tokens_per_kvp_group
                % self._replica_config.block_size
                == 0
            ), "max_num_tokens_per_kvp_group should be a multiple of block_size"

            self._max_num_tokens_per_kvp_group = (
                self._replica_config.max_num_tokens_per_kvp_group
            )
            self._max_num_blocks_per_kvp_group = ceil(
                self._max_num_tokens_per_kvp_group / self._replica_config.block_size
            )

            assert (
                self._config.num_blocks > self._max_num_blocks_per_kvp_group
            ), "num_blocks should be greater than max_num_blocks_per_kvp_group"
        else:
            self._max_num_tokens_per_kvp_group = min(
                self._request_generator_config.max_tokens,
                self._config.num_blocks * self._replica_config.block_size,
            )
            self._max_num_blocks_per_kvp_group = min(
                self._max_blocks_per_sequence, self._config.num_blocks
            )

        # kvp_group_id -> request_id -> num_blocks
        self._kvp_group_allocation_map: List[Dict[int, int]] = [
            defaultdict(int) for _ in range(self._kv_parallel_size)
        ]
        # kvp_group_id -> num_allocated_blocks
        self._kvp_group_num_allocated_blocks: List[int, int] = [
            0 for _ in range(self._kv_parallel_size)
        ]
        # request_id -> kvp_group_id -> num_blocks
        self._req_kvp_group_block_counter: Dict[int, OrderedDict[int, int]] = {}
        # request_id -> num_blocks
        self._req_block_counter: Dict[int, int] = defaultdict(int)
        # kvp_group -> pending prefill work
        self._kvp_group_pending_prefill_work: List[int] = [
            0 for _ in range(self._kv_parallel_size)
        ]

    def _get_batch_formation_tracker(self) -> BatchFormationTracker:
        return BatchFormationTracker(
            self._replica_id,
            self._max_micro_batch_size,
            self._kv_parallel_size,
            self._max_num_tokens_per_kvp_group,
        )

    def add_request(self, request):
        return self._request_queue.append(
            (self._get_request_priority(request), request)
        )

    def _can_allocate_blocks(self, kvp_group_id: int, num_blocks: int) -> bool:
        return (
            self._config.num_blocks
            - self._kvp_group_num_allocated_blocks[kvp_group_id]
            - num_blocks
            >= self._watermark_blocks
        )

    def allocate(self, request_id: int, num_blocks: int, kvp_group_id: int) -> None:
        self._kvp_group_allocation_map[kvp_group_id][request_id] += num_blocks

        if not request_id in self._req_kvp_group_block_counter:
            self._req_kvp_group_block_counter[request_id] = OrderedDict()

        if not kvp_group_id in self._req_kvp_group_block_counter[request_id]:
            self._req_kvp_group_block_counter[request_id][kvp_group_id] = 0

        self._req_kvp_group_block_counter[request_id][kvp_group_id] += num_blocks
        self._kvp_group_num_allocated_blocks[kvp_group_id] += num_blocks

        assert (
            self._kvp_group_num_allocated_blocks[kvp_group_id]
            <= self._config.num_blocks
        )

        self._req_block_counter[request_id] += num_blocks

    def _get_allocation_order(self) -> List[int]:
        """
        We are simply picking the kvp group with the least prefill work.
        We use the square of number of prefill tokens as a simple proxy for prefill work of now.

        TODO(amey): since the initial chunks are linearly proportional to the number of tokens,
        this can lead to some inbalance in the allocation. We can improve this in the future.
        """
        return sorted(
            range(self._kv_parallel_size),
            key=lambda kvp_group_id: self._kvp_group_pending_prefill_work[kvp_group_id],
        )

    def _allocate_request(self, request: Request) -> None:
        """
        We use a naive approach to allocate memory where we allocate all the memory
        required by the request in one go. This is because we expect the compute requirement
        to far exceed the memory requirement. In KVP, incremental memory allocation can
        lead to deadlocks -- where multiple long requests are waiting for memory to be available
        on a new kvp group, but none of them can proceed because the memory is not available.

        TODO(amey): This is a naive approach and can be improved in the future. Especially, offloading
        memory allocation to CPU can be a good solution, especially for longer requests.

        While allocating memory, we must choose the kvp groups such that we have minimal
        compute contention. While also ensuring that we don't create memory hotspots.
        The allocate method offloads this responsibility to _get_allocation_order method.
        """
        if request.id in self._req_block_counter:
            return True

        if (
            request.num_prefill_tokens
            > self._kv_parallel_size * self._max_num_tokens_per_kvp_group
        ):
            logger.warn(
                f"Request {request.id} requires {request.num_prefill_tokens} tokens, which is more the kv parallel size {self._kv_parallel_size} * max num tokens per kvp group {self._max_num_tokens_per_kvp_group}"
            )
            return False

        num_required_blocks = ceil(
            request.num_prefill_tokens / self._replica_config.block_size
        )

        # more than one group is required
        num_kvp_groups = ceil(num_required_blocks / self._max_num_blocks_per_kvp_group)
        num_kvp_groups = min(num_kvp_groups, self._kv_parallel_size)
        last_group_num_blocks = (
            num_required_blocks
            - self._max_num_blocks_per_kvp_group * (num_kvp_groups - 1)
        )

        if num_kvp_groups == 1:
            for kvp_group_id in self._get_allocation_order():
                if self._can_allocate_blocks(kvp_group_id, num_required_blocks):
                    self.allocate(request.id, num_required_blocks, kvp_group_id)
                    return True
            return False

        num_kvp_groups_found = 0
        last_kvp_group_found = False

        kvp_group_ids: List[int] = []
        last_kvp_group_id: Optional[int] = None

        # check if any combination of workers can accommodate this sequence
        for kvp_group_id in self._get_allocation_order():
            """
            This loop represents a simple first-fit allocation strategy.
            """
            if self._can_allocate_blocks(
                kvp_group_id, self._max_num_blocks_per_kvp_group
            ):
                num_kvp_groups_found += 1
                kvp_group_ids.append(kvp_group_id)
            elif (
                last_group_num_blocks
                and not last_kvp_group_found
                and self._can_allocate_blocks(kvp_group_id, last_group_num_blocks)
            ):
                last_kvp_group_found = True
                num_kvp_groups_found += 1
                last_kvp_group_id = kvp_group_id

            if num_kvp_groups_found == num_kvp_groups:
                break

        if num_kvp_groups_found != num_kvp_groups:
            return False

        if last_kvp_group_id:
            kvp_group_ids.append(last_kvp_group_id)
        else:
            last_kvp_group_id = kvp_group_ids[-1]

        for kvp_group_id in kvp_group_ids:
            if kvp_group_id == last_kvp_group_id:
                self.allocate(request.id, last_group_num_blocks, kvp_group_id)
            else:
                self.allocate(
                    request.id, self._max_num_blocks_per_kvp_group, kvp_group_id
                )

            # TODO(amey): right now we are adding same weight to all the kvp groups
            # in reality later groups have lower load, also the load will be delayed
            # so we should have more work added to the earlier groups
            self._kvp_group_pending_prefill_work[kvp_group_id] += (
                request.num_prefill_tokens**2
            )

        return True

    def free(self, *request_ids: List[int]) -> None:
        for request_id in request_ids:
            for kvp_group_id in self._req_kvp_group_block_counter[request_id]:
                num_blocks = self._req_kvp_group_block_counter[request_id][kvp_group_id]
                self._kvp_group_num_allocated_blocks[kvp_group_id] -= num_blocks
                self._kvp_group_allocation_map[kvp_group_id].pop(request_id)

                assert self._kvp_group_num_allocated_blocks[kvp_group_id] >= 0

            self._req_block_counter.pop(request_id)
            self._req_kvp_group_block_counter.pop(request_id)

    def _can_append_slot(self, request: Request) -> bool:
        last_kvp_group_id = list(self._req_kvp_group_block_counter[request.id].keys())[
            -1
        ]
        return (
            self._kvp_group_num_allocated_blocks[last_kvp_group_id]
            < self._config.num_blocks
        )

    def _append_slot(self, request: Request) -> None:
        num_tokens_reserved = (
            self._req_block_counter[request.id] * self._replica_config.block_size
        )
        num_tokens_required = max(0, request.num_processed_tokens - num_tokens_reserved)
        if num_tokens_required == 0:
            return

        last_kvp_group_id = list(self._req_kvp_group_block_counter[request.id].keys())[
            -1
        ]
        self.allocate(request.id, 1, last_kvp_group_id)

    def on_batch_end(self, batch: KVParallelBatch) -> None:
        super().on_batch_end(batch)

        for request in batch.requests:
            if request.completed:
                self.free(request.id)
                continue

            if not self.is_sequence_pipeline_parallel_enabled:
                if not request.is_prefill_complete:
                    self.add_request(request)
                else:
                    self._preempted_requests.append(request)
            else:
                if request.is_prefill_complete:
                    self._preempted_requests.append(request)

        for kvp_group_id, sub_batch in batch.batch_mapping.items():
            for request in sub_batch.requests:
                # we need to update the prefill work for the kvp groups
                if request.num_processed_decode_tokens > 1:
                    continue

                total_work = request.num_prefill_tokens**2
                completed_work = self._get_num_processed_tokens(request) ** 2
                self._kvp_group_pending_prefill_work[kvp_group_id] = (
                    total_work - completed_work
                )

    def on_stage_end(
        self, stage_id: int, batch_stage: Union[BatchStage, KVParallelBatchStage]
    ) -> None:
        super().on_stage_end(stage_id, batch_stage)

        if not stage_id == 0:
            return

        if not self.is_sequence_pipeline_parallel_enabled:
            return

        for request in batch_stage.requests:
            assert not request.completed

            if request.is_prefill_stage_complete:
                continue

            self.add_request(request)

    def _is_prefill_complete(self, request: Request) -> bool:
        if self.is_sequence_pipeline_parallel_enabled:
            return request.is_prefill_stage_complete
        else:
            return request.is_prefill_complete

    def _get_num_processed_tokens(self, request: Request) -> int:
        if self.is_sequence_pipeline_parallel_enabled:
            if not request.is_prefill_stage_complete:
                return request.num_stage_processed_tokens

            return request.num_processed_tokens
        else:
            assert not request.is_prefill_complete
            return request.num_processed_tokens

    @abstractmethod
    def _get_request_next_num_q_tokens(
        self,
        time: float,
        request: Request,
        batch_formation_tracker: BatchFormationTracker,
    ) -> int:
        pass

    def get_pending_requests(self) -> List[Request]:
        return [x[1] for x in self._request_queue]

    def _get_active_kvp_group_ids(self, request: Request) -> List[int]:
        kvp_group_ids = list(self._req_kvp_group_block_counter[request.id].keys())

        if self._is_prefill_complete(request):
            return kvp_group_ids

        num_kvp_groups = (
            self._get_num_processed_tokens(request)
            // self._max_num_tokens_per_kvp_group
            + 1
        )

        return kvp_group_ids[:num_kvp_groups]

    def _ensure_can_append_slot(self, request: Request) -> bool:
        if self._can_append_slot(request):
            return True

        could_ensure_memory = False

        while not self._can_append_slot(request):
            if self._preempted_requests:
                last_kvp_group_id = list(
                    self._req_kvp_group_block_counter[request.id].keys()
                )[-1]
                # find the last request that contains allocation on the last kv group

                # TODO(amey): here we are just restarting the request based on the scheduling preference order
                # however, this doesn't account for the size of the request -- so potentially we could be
                # restarting a large request instead of a smaller one. We can improve this in the future.
                for i, req in enumerate(reversed(self._preempted_requests)):
                    if (
                        not last_kvp_group_id
                        in self._req_kvp_group_block_counter[req.id]
                    ):
                        continue

                    victim_request = self._preempted_requests.pop(-1 * (i + 1))
                    victim_request.restart()
                    self.free(victim_request.id)
                    self.add_request(victim_request)
                    could_ensure_memory = True
                    break
            else:
                request.restart()
                self.free(request.id)
                self.add_request(request)
                could_ensure_memory = False
                break

        return could_ensure_memory

    @abstractmethod
    def _get_request_priority(self, request: Request) -> Any:
        pass

    def _sort_request_queue(self, time: float) -> None:
        self._request_queue.sort(key=lambda x: x[0])

    def _get_next_batch(self, time: float) -> KVParallelBatch:
        batch_formation_tracker = self._get_batch_formation_tracker()

        # preempted requests could contain multiple requests which have
        # partial prefills completed, so we need to be careful
        while self._preempted_requests:
            if not batch_formation_tracker.can_add_request():
                break

            request = self._preempted_requests.pop(0)

            assert self._is_prefill_complete(request)

            can_append_slot = self._ensure_can_append_slot(request)

            if not can_append_slot:
                # the request has been restarted and added to the wait queue
                continue

            self._append_slot(request)
            active_kvp_groups = self._get_active_kvp_group_ids(request)
            num_processed_tokens = self._get_num_processed_tokens(request)
            batch_formation_tracker.add_request(
                request, 1, num_processed_tokens, active_kvp_groups
            )

        self._sort_request_queue(time)

        num_skipped_requests = 0

        while (
            len(self._request_queue) > num_skipped_requests
            and num_skipped_requests < MAX_NUM_SKIPPED_REQUESTS
        ):
            _, request = self._request_queue[num_skipped_requests]

            is_new_request = not request.id in self._req_block_counter

            if (
                is_new_request
                and len(self._req_block_counter) == self._config.batch_size_cap
            ):
                num_skipped_requests += 1
                continue

            assert not self._is_prefill_complete(request)

            if not batch_formation_tracker.can_add_request():
                num_skipped_requests += 1
                # print(f"Could not add request {request.id} to batch")
                continue

            if not self._allocate_request(request):
                num_skipped_requests += 1
                # print(f"Could not allocate request {request.id}")
                continue

            next_num_q_tokens = self._get_request_next_num_q_tokens(
                time, request, batch_formation_tracker
            )

            if next_num_q_tokens == 0:
                num_skipped_requests += 1
                # print(f"Request {request.id} has 0 q tokens")
                continue

            self._request_queue.pop(num_skipped_requests)

            active_kvp_groups = self._get_active_kvp_group_ids(request)
            num_processed_tokens = self._get_num_processed_tokens(request)
            batch_formation_tracker.add_request(
                request, next_num_q_tokens, num_processed_tokens, active_kvp_groups
            )

        batch = batch_formation_tracker.get_batch()

        # print(f"Scheduling Batch: {batch}")

        return batch



from vidur.scheduler.global_scheduler.lor_global_scheduler import LORGlobalScheduler
from vidur.scheduler.global_scheduler.random_global_scheduler import (
    RandomGlobalScheduler,
)
from vidur.scheduler.global_scheduler.round_robin_global_scheduler import (
    RoundRobinGlobalScheduler,
)
from vidur.types import GlobalSchedulerType
from vidur.utils.base_registry import BaseRegistry


class GlobalSchedulerRegistry(BaseRegistry):
    @classmethod
    def get_key_from_str(cls, key_str: str) -> GlobalSchedulerType:
        return GlobalSchedulerType.from_str(key_str)


GlobalSchedulerRegistry.register(GlobalSchedulerType.RANDOM, RandomGlobalScheduler)
GlobalSchedulerRegistry.register(
    GlobalSchedulerType.ROUND_ROBIN, RoundRobinGlobalScheduler
)
GlobalSchedulerRegistry.register(GlobalSchedulerType.LOR, LORGlobalScheduler)



import os
import subprocess
import sys
import shutil
from typing import List
from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext
from setuptools_scm import get_version


ROOT_DIR = os.path.dirname(__file__)


def is_ninja_available() -> bool:
    return shutil.which("ninja") is not None


def get_path(*filepath) -> str:
    return os.path.join(ROOT_DIR, *filepath)


def get_requirements() -> List[str]:
    """Get Python package dependencies from requirements.txt."""
    with open(get_path("requirements.txt")) as f:
        requirements = f.read().strip().split("\n")
    return requirements


class CMakeExtension(Extension):
    def __init__(self, name: str, cmake_lists_dir: str = '.', **kwa) -> None:
        super().__init__(name, sources=[], **kwa)
        self.cmake_lists_dir = os.path.abspath(cmake_lists_dir)


class CMakeBuildExt(build_ext):
    did_config = {}

    def compute_num_jobs(self):
        try:
            num_jobs = len(os.sched_getaffinity(0))
        except AttributeError:
            num_jobs = os.cpu_count()
        return num_jobs

    def configure(self, ext: CMakeExtension) -> None:
        # If we've already configured using the CMakeLists.txt for
        # this extension, exit early.
        if ext.cmake_lists_dir in CMakeBuildExt.did_config:
            return

        CMakeBuildExt.did_config[ext.cmake_lists_dir] = True

        default_cfg = "Debug" if self.debug else "RelWithDebInfo"
        cfg = os.getenv("CMAKE_BUILD_TYPE", default_cfg)

        # Where .so files will be written
        outdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

        cmake_args = [
            f'-DCMAKE_BUILD_TYPE={cfg}',
            f'-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={outdir}',
            f'-DVIDUR_PYTHON_EXECUTABLE={sys.executable}',
        ]

        verbose = bool(int(os.getenv('VERBOSE', '0')))
        if verbose:
            cmake_args += ['-DCMAKE_VERBOSE_MAKEFILE=ON']

        # Setup build tool
        if is_ninja_available():
            num_jobs = self.compute_num_jobs()
            build_tool = ['-G', 'Ninja']
            cmake_args += [
                '-DCMAKE_JOB_POOL_COMPILE:STRING=compile',
                f'-DCMAKE_JOB_POOLS:STRING=compile={num_jobs}',
            ]
        else:
            build_tool = []

        subprocess.check_call(
            ['cmake', ext.cmake_lists_dir, *build_tool, *cmake_args],
            cwd=self.build_temp)

        include_source = os.path.join(ext.cmake_lists_dir, 'csrc', 'include')
        include_destination = os.path.join(outdir, 'include')
        shutil.copytree(include_source, include_destination)

    def build_extensions(self) -> None:
        # Ensure CMake is present
        try:
            subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError('Cannot find CMake executable')

       # Create and clean build directory
        if os.path.exists(self.build_temp):
            shutil.rmtree(self.build_temp)  # Clean up any existing build artifacts
        os.makedirs(self.build_temp)

        # Build extensions
        for ext in self.extensions:
            self.configure(ext)
            num_jobs = self.compute_num_jobs()
            
            # Build both targets
            build_args = ['--build', '.', '--target', '_native', '-j', str(num_jobs)]
            subprocess.check_call(['cmake', *build_args], cwd=self.build_temp)


# Define extensions
ext_modules = [CMakeExtension(name="vidur._native")]

# include_files = []
# for root, dirs, files in os.walk('csrc/include'):
#     for file in files:
#         include_files.append(os.path.join(root, file))


setup(
    author="Systems for AI Lab, Georgia Tech; Microsoft Corporation",
    python_requires='>=3.10',
    description="A LLM inference cluster simulator",
    keywords='Simulator, LLM, Inference, Cluster',
    name='vidur',
    packages=find_packages(include=['vidur', 'vidur.*']),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://mlsys.org/virtual/2024/poster/2667',
    version=get_version(),
    install_requires=get_requirements(),
    ext_modules=ext_modules,
    cmdclass={'build_ext': CMakeBuildExt},
    entry_points={
        "console_scripts": [
            "vidur-data = vidur.cli.data_cli:cli",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import hashlib
import json
import os
import pickle
from abc import abstractmethod
from enum import Enum
from itertools import product
from typing import Any, Dict, List, Tuple

import numpy as np
import pandas as pd
from fasteners import InterProcessReaderWriterLock
from huggingface_hub import snapshot_download
from sklearn.base import BaseEstimator
from sklearn.metrics import make_scorer
from sklearn.model_selection import GridSearchCV

from vidur._native.config import (
    ExecutionTimePredictorConfig as ExecutionTimePredictorConfigC,
)
from vidur._native.config import ModelConfig as ModelConfigC
from vidur._native.config import ReplicaConfig as ReplicaConfigC
from vidur._native.execution_time_predictor import (
    ExecutionTimePredictor as ExecutionTimePredictorC,
)
from vidur.config import BaseExecutionTimePredictorConfig, ReplicaConfig
from vidur.entities import Batch
from vidur.execution_time_predictor.base_execution_time_predictor import (
    BaseExecutionTimePredictor,
)
from vidur.logger import init_logger
from vidur.utils import sanitize_name
from vidur.utils.hf_dataset_utils import upload_dataset
from vidur.utils.parallel import parallel_map

logger = init_logger(__name__)


class ModelType(Enum):
    COMPUTE = "compute"
    NETWORK = "network"


def _store_prediction_file(file_path: str, predictions: Dict[Tuple, float]) -> None:
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    df = pd.DataFrame(
        [(*k, v) for k, v in predictions.items()],
        columns=[f"X_{i}" for i in range(len(next(iter(predictions))))]
        + ["prediction"],
    )
    df.to_csv(file_path, index=False, compression="xz")


def _load_prediction_file(file_path: str) -> Dict[Tuple, float]:
    """
    Load with tuple keys
    """
    df = pd.read_csv(file_path)
    pred_col = df.pop("prediction")
    return {tuple(row): pred for row, pred in zip(df.values, pred_col)}


class SklearnExecutionTimePredictor(BaseExecutionTimePredictor):
    def __init__(
        self,
        predictor_config: BaseExecutionTimePredictorConfig,
        replica_config: ReplicaConfig,
    ) -> None:
        super().__init__(
            predictor_config=predictor_config,
            replica_config=replica_config,
        )
        os.makedirs(self._cache_dir, exist_ok=True)

        # These overheads are only for GQA models
        self._attention_decode_batching_overhead_fraction = (
            (self._config.attention_decode_batching_overhead_fraction)
            if self._model_config.num_q_heads > self._model_config.num_kv_heads
            else 0
        )
        # TODO(amey): this won't work for orca scheduler
        self._max_tokens = self._config.prediction_max_tokens_per_request

        num_workers = (
            self._replica_config.num_pipeline_stages
            * self._replica_config.tensor_parallel_size
        )
        devices_per_node = self._replica_config.node_config.num_devices_per_node
        assert (
            num_workers < devices_per_node or num_workers % devices_per_node == 0
        ), "Number of workers should be less than devices per node or a multiple of devices per node"

        self._is_multi_node = num_workers > devices_per_node

        loaded_predictions_from_hf = False
        try:
            self._predictions = self._load_predictions_from_hf()
            loaded_predictions_from_hf = True
        except Exception as e:
            logger.error(f"Failed to load predictions from HF: {e}")

        if not loaded_predictions_from_hf:
            models = self._train_models()
            self._predictions = self._predict_from_models(models)

            if self._config.upload_predictions and not self._config.no_cache:
                assert self._config.prediction_hf_org is not None
                assert self._config.prediction_hf_collection is not None
                self._upload_predictions_to_hf()

        if self._config.use_native_execution_time_predictor:
            self._native_execution_time_predictor = (
                self._init_native_execution_time_predictor()
            )

    def _init_native_execution_time_predictor(self) -> None:
        predictor_config = ExecutionTimePredictorConfigC(
            kv_cache_prediction_granularity=self._config.kv_cache_prediction_granularity,
            prediction_max_prefill_chunk_size=self._config.prediction_max_prefill_chunk_size,
            prediction_max_batch_size=self._config.prediction_max_batch_size,
            prediction_max_tokens_per_request=self._config.prediction_max_tokens_per_request,
            attention_decode_batching_overhead_fraction=self._config.attention_decode_batching_overhead_fraction,
            nccl_cpu_launch_overhead_ms=self._config.nccl_cpu_launch_overhead_ms,
            nccl_cpu_skew_overhead_per_device_ms=self._config.nccl_cpu_skew_overhead_per_device_ms,
            use_native_execution_time_predictor=self._config.use_native_execution_time_predictor,
            disable_kvp_communication=self._config.disable_kvp_communication,
        )

        replica_config = ReplicaConfigC(
            num_pipeline_stages=self._replica_config.num_pipeline_stages,
            tensor_parallel_size=self._replica_config.tensor_parallel_size,
            kv_parallel_size=self._replica_config.kv_parallel_size,
        )

        model_config = ModelConfigC(
            num_layers=self._replica_config.model_config.num_layers,
            num_q_heads=self._replica_config.model_config.num_q_heads,
            num_kv_heads=self._replica_config.model_config.num_kv_heads,
            embedding_dim=self._replica_config.model_config.embedding_dim,
            mlp_hidden_dim=self._replica_config.model_config.mlp_hidden_dim,
            max_model_len=self._replica_config.model_config.max_model_len,
            use_gated_mlp=self._replica_config.model_config.use_gated_mlp,
            use_bias=self._replica_config.model_config.use_bias,
            use_qkv_bias=self._replica_config.model_config.use_qkv_bias,
            post_attn_norm=self._replica_config.model_config.post_attn_norm,
            vocab_size=self._replica_config.model_config.vocab_size,
        )

        prediction_ops = []
        prediction_keys = []
        prediction_values = []

        for prediction_op, prediction_dict in self._predictions.items():
            prediction_ops.append(prediction_op)
            prediction_keys.append([])
            prediction_values.append([])
            for key, value in prediction_dict.items():
                if len(key) == 1:
                    prediction_keys[-1].append((int(key[0]), -1))
                elif len(key) == 2:
                    prediction_keys[-1].append((int(key[0]), int(key[1])))
                else:
                    raise Exception(f"Key has more than 2 dimensions: {type(key)}")
                prediction_values[-1].append(value)

        self._native_execution_time_predictor = ExecutionTimePredictorC(
            config=predictor_config,
            replica_config=replica_config,
            model_config=model_config,
            prediction_keys=prediction_keys,
            prediction_values=prediction_values,
            prediction_ops=prediction_ops,
        )

        return self._native_execution_time_predictor

    def _get_input_files(self) -> Tuple[str, str, str, str, str]:
        # Download the compute and network profile dataset from the huggingface datasets
        # specified in the config into cache directory
        compute_dataset_path = snapshot_download(
            repo_id=self._config.model_profiling_dataset,
            repo_type="dataset",
            cache_dir=f"{self._cache_dir}/profiling_data/compute/",
        )

        logger.info(f"🔽 Downloaded compute data to: {compute_dataset_path}")

        compute_dataset_config = json.load(
            open(f"{compute_dataset_path}/config.json", "r")
        )

        assert compute_dataset_config["model"] == self._replica_config.model_name
        assert compute_dataset_config["device_sku"] == self._replica_config.device

        network_dataset_path = snapshot_download(
            repo_id=self._config.network_profiling_dataset,
            repo_type="dataset",
            cache_dir=f"{self._cache_dir}/profiling_data/network/",
        )

        logger.info(f"🔽 Downloaded network data to: {network_dataset_path}")

        network_dataset_config = json.load(
            open(f"{network_dataset_path}/config.json", "r")
        )

        assert network_dataset_config["sku"] == self._replica_config.network_device

        return (
            f"{compute_dataset_path}/mlp.csv.xz",
            f"{compute_dataset_path}/attention.csv.xz",
            f"{network_dataset_path}/all_reduce.csv.xz",
            f"{network_dataset_path}/send_recv.csv.xz",
        )

    def _load_compute_df(self, file_path: str) -> pd.DataFrame:
        df = self._read_input_file(file_path)
        df = df.drop_duplicates()

        logger.debug(f"Length of complete compute df: {len(df)} {file_path}")
        logger.debug(f"num_q_heads: {self._model_config.num_q_heads}")
        logger.debug(f"embedding_dim: {self._model_config.embedding_dim}")
        logger.debug(f"mlp_hidden_dim: {self._model_config.mlp_hidden_dim}")
        logger.debug(f"use_gated_mlp: {self._model_config.use_gated_mlp}")
        logger.debug(f"vocab_size: {self._model_config.vocab_size}")
        logger.debug(
            f"num_tensor_parallel_workers: {self._replica_config.tensor_parallel_size}"
        )

        if (
            self._replica_config.tensor_parallel_size
            not in df["num_tensor_parallel_workers"].unique()
        ):
            raise Exception(
                f"Tensor parallel size {self._replica_config.tensor_parallel_size} not found in profile data: {file_path}\nAvailable TP sizes include:",
                df["num_tensor_parallel_workers"].unique(),
            )

        df = df[
            (df["n_head"] == self._model_config.num_q_heads)
            & (df["n_kv_head"] == self._model_config.num_kv_heads)
            & (df["n_embd"] == self._model_config.embedding_dim)
            & (df["n_expanded_embd"] == self._model_config.mlp_hidden_dim)
            & (df["use_gated_mlp"] == self._model_config.use_gated_mlp)
            & (df["vocab_size"] == self._model_config.vocab_size)
            & (
                df["num_tensor_parallel_workers"]
                == self._replica_config.tensor_parallel_size
            )
        ]

        for column in [
            "time_stats.post_attention_layernorm.median",
            "time_stats.add.median",
            "time_stats.input_layernorm.median",
        ]:
            if column not in df.columns:
                df[column] = 0
            else:
                df.fillna({column: 0}, inplace=True)
        return df

    def _load_attention_df(self, file_path: str) -> pd.DataFrame:
        df = pd.read_csv(file_path)
        df = df.drop_duplicates()

        for column in [
            "time_stats.attn_kv_cache_save.median",
        ]:
            if column not in df.columns:
                df[column] = 0
            else:
                df.fillna({column: 0}, inplace=True)

        if (
            self._replica_config.tensor_parallel_size
            not in df["num_tensor_parallel_workers"].unique()
        ):
            raise Exception(
                f"Tensor parallel size {self._replica_config.tensor_parallel_size} not found in profile data: {file_path}\nAvailable TP sizes include:",
                df["num_tensor_parallel_workers"].unique(),
            )

        return df[
            (df["n_embd"] == self._model_config.embedding_dim)
            & (df["n_q_head"] == self._model_config.num_q_heads)
            & (df["n_kv_head"] == self._model_config.num_kv_heads)
            & (df["block_size"] == self._block_size)
            & (
                df["num_tensor_parallel_workers"]
                == self._replica_config.tensor_parallel_size
            )
        ]

    def _load_all_reduce_df(self, file_path: str) -> pd.DataFrame:
        df = self._read_input_file(file_path)

        if self._replica_config.tensor_parallel_size not in df["num_workers"].unique():
            raise Exception(
                f"Tensor parallel size {self._replica_config.tensor_parallel_size} not found in profile data: {file_path}\nAvailable TP sizes include:",
                df["num_workers"].unique(),
            )

        return df[
            (df["num_workers"] == self._replica_config.tensor_parallel_size)
            & (df["devices_per_node"] == self._replica_config.tensor_parallel_size)
            & (df["collective"] == "all_reduce")
        ]

    def _load_all_reduce_kvp_df(self, file_path: str) -> pd.DataFrame:
        df = self._read_input_file(file_path)
        # TODO(amey): We are assuming that KVP workers are always going to be staggered
        return df[(df["devices_per_node"] == 1) & (df["collective"] == "all_reduce")]

    def _load_send_recv_df(self, file_path: str) -> pd.DataFrame:
        # TODO(Amey): We are making a simplifying assumption here that if
        # we go cross node in the model, pp communication is going to happen
        # over inter-node network. This is not always true of course.
        if self._is_multi_node:
            devices_per_node = 1
        else:
            devices_per_node = 2

        df = self._read_input_file(file_path)
        filtered_df = df[
            (df["collective"] == "send_recv")
            & (df["devices_per_node"] == devices_per_node)
        ]
        return filtered_df

    def _read_input_file(self, file_path: str) -> pd.DataFrame:
        df = pd.read_csv(file_path)
        df = df.drop_duplicates()
        return df

    def _get_compute_df_with_derived_features(self, df: pd.DataFrame) -> pd.DataFrame:
        df_with_derived_features = df.copy()
        return df_with_derived_features

    def _get_attention_df_with_derived_features(self, df: pd.DataFrame) -> pd.DataFrame:
        df_with_derived_features = df.copy()
        df_with_derived_features["num_tokens"] = df_with_derived_features[
            ["prefill_chunk_size", "batch_size"]
        ].max(axis=1)
        df_with_derived_features["is_decode"] = (
            df_with_derived_features["prefill_chunk_size"] == 0
        )
        df_with_derived_features["prefill_chunk_size_squared"] = (
            df_with_derived_features["prefill_chunk_size"] ** 2
        )
        return df_with_derived_features

    def _get_all_reduce_df_with_derived_features(
        self, df: pd.DataFrame
    ) -> pd.DataFrame:
        df_with_derived_features = df.copy()
        # convert bytes to num tokens
        # each token is of size 2 * h bytes
        df_with_derived_features["num_tokens"] = (
            df_with_derived_features["size"] / self._model_config.embedding_dim / 2
        )
        return df_with_derived_features

    def _get_send_recv_df_with_derived_features(self, df: pd.DataFrame) -> pd.DataFrame:
        df_with_derived_features = df.copy()
        df_with_derived_features["num_tokens"] = (
            df_with_derived_features["size"] / self._model_config.embedding_dim / 2
        )
        return df_with_derived_features

    def _get_compute_prediction_repo_id(self) -> str:
        model_hash = sanitize_name(self._get_hash(ModelType.COMPUTE))
        # Create standardized repo identifier
        org = sanitize_name(self._config.prediction_hf_org)
        collection = sanitize_name(self._config.prediction_hf_collection)
        profile_dataset_name = sanitize_name(self._config.model_profiling_dataset)

        repo_id = f"{org}/predictions-{collection}-{profile_dataset_name}-{model_hash}"

        return repo_id

    def _get_network_prediction_repo_id(self) -> str:
        model_hash = sanitize_name(self._get_hash(ModelType.NETWORK))
        # Create standardized repo identifier
        org = sanitize_name(self._config.prediction_hf_org)
        collection = sanitize_name(self._config.prediction_hf_collection)
        profile_dataset_name = sanitize_name(self._config.network_profiling_dataset)

        repo_id = f"{org}/predictions-{collection}-{profile_dataset_name}-{model_hash}"

        return repo_id

    def _load_predictions_from_hf(self) -> bool:
        compute_repo_id = self._get_compute_prediction_repo_id()
        network_repo_id = self._get_network_prediction_repo_id()

        # try to download the predictions from HF
        compute_output_dir = snapshot_download(
            repo_id=compute_repo_id,
            repo_type="dataset",
            cache_dir=f"{self._cache_dir}/hf_predictions/compute/",
        )
        logger.info(f"✅ Downloaded compute predictions from: {compute_repo_id}")

        network_output_dir = snapshot_download(
            repo_id=network_repo_id,
            repo_type="dataset",
            cache_dir=f"{self._cache_dir}/hf_predictions/network/",
        )
        logger.info(f"✅ Downloaded network predictions from: {network_repo_id}")

        all_prediction_files: List[str] = []
        model_names: List[str] = []
        for output_dir in [compute_output_dir, network_output_dir]:
            for file in os.listdir(output_dir):
                if file.endswith(".csv.xz"):
                    all_prediction_files.append(f"{output_dir}/{file}")
                    model_names.append(file.split(".")[0])

        predictions_dicts = parallel_map(_load_prediction_file, all_prediction_files)

        return {
            model_name: predictions
            for model_name, predictions in zip(model_names, predictions_dicts)
        }

    def _upload_predictions_to_hf(self) -> None:
        compute_repo_id = self._get_compute_prediction_repo_id()
        network_repo_id = self._get_network_prediction_repo_id()

        # Create config.json
        compute_model_config = {
            "schema_version": "1.0",
            **self._get_compute_model_config(),
        }

        upload_dataset(
            repo_id=compute_repo_id,
            config=compute_model_config,
            data_path=f"{self._cache_dir}/predictions/{self._get_hash(ModelType.COMPUTE)}",
            use_private_repo=self._config.upload_to_private_repo,
        )

        logger.info(f"✅ Published compute predictions to: {compute_repo_id}")

        network_model_config = {
            "schema_version": "1.0",
            **self._get_network_model_config(),
        }

        upload_dataset(
            repo_id=network_repo_id,
            config=network_model_config,
            data_path=f"{self._cache_dir}/predictions/{self._get_hash(ModelType.NETWORK)}",
            use_private_repo=self._config.upload_to_private_repo,
        )

        logger.info(f"✅ Published network predictions to: {network_repo_id}")

    @staticmethod
    def mean_absolute_percentage_error(y_true: np.array, y_pred: np.array) -> float:
        y_true, y_pred = np.array(y_true), np.array(y_pred)
        # Handling the case where y_true is 0 separately to avoid division by zero
        zero_true_mask = y_true == 0
        non_zero_true_mask = ~zero_true_mask

        # For non-zero true values, calculate the absolute percentage error
        error = np.zeros_like(y_true, dtype=float)  # using float instead of np.float
        error[non_zero_true_mask] = (
            np.abs(
                (y_true[non_zero_true_mask] - y_pred[non_zero_true_mask])
                / y_true[non_zero_true_mask]
            )
            * 100
        )

        # For zero true values, if prediction is also 0, error is 0, else it is 100
        error[zero_true_mask] = np.where(y_pred[zero_true_mask] == 0, 0, 100)

        # Return the mean of the absolute percentage errors
        return np.mean(error)

    def _get_scorer(self) -> Any:
        return make_scorer(
            SklearnExecutionTimePredictor.mean_absolute_percentage_error,
            greater_is_better=False,
        )

    @abstractmethod
    def _get_grid_search_params(self) -> Dict[str, Any]:
        pass

    @abstractmethod
    def _get_estimator(self) -> BaseEstimator:
        pass

    def _get_hash(self, model_type: ModelType) -> str:
        if model_type == ModelType.COMPUTE:
            config_str = str(self._get_compute_model_config())
        elif model_type == ModelType.NETWORK:
            config_str = str(self._get_network_model_config())
        else:
            raise ValueError(f"Invalid model type: {model_type}")

        return hashlib.md5(config_str.encode("utf-8")).hexdigest()[0:8]

    def _load_model_from_cache(self, model_name: str, model_hash: str) -> BaseEstimator:
        with InterProcessReaderWriterLock(
            f"{self._cache_dir}/{model_name}_{model_hash}_model_lock.file"
        ).read_lock():
            if self._config.no_cache:
                return
            # check if model is in cache
            cache_file = f"{self._cache_dir}/predictors/{model_hash}/{model_name}.pkl"

            if not os.path.exists(cache_file):
                return

            logger.debug(f"Found model {model_name} in cache")
            model = pickle.load(open(cache_file, "rb"))
            return model

    def _store_model_in_cache(
        self, model_name: str, model_hash: str, model: BaseEstimator
    ) -> None:
        with InterProcessReaderWriterLock(
            f"{self._cache_dir}/{model_hash}_model_lock.file"
        ).write_lock():
            # store model in cache
            cache_file = f"{self._cache_dir}/predictors/{model_hash}/{model_name}.pkl"
            os.makedirs(os.path.dirname(cache_file), exist_ok=True)
            pickle.dump(model, open(cache_file, "wb"), protocol=pickle.HIGHEST_PROTOCOL)

    def _store_training_prediction_data(
        self,
        model_name: str,
        model_hash: str,
        df: pd.DataFrame,
        feature_cols: List[str],
        target_col: str,
        model: BaseEstimator,
    ) -> None:
        df = df.copy()

        # convert the df to list of tuples
        df["prediction"] = model.predict(df[feature_cols])

        # store the prediction data
        output_file = (
            f"{self._cache_dir}/training_predictions/{model_hash}/{model_name}.csv"
        )

        os.makedirs(os.path.dirname(output_file), exist_ok=True)

        df[feature_cols + [target_col, "prediction"]].to_csv(
            output_file,
            index=False,
        )

    def _train_model(
        self,
        model_name: str,
        model_type: ModelType,
        df: pd.DataFrame,
        feature_cols: List[str],
        target_col: str,
    ) -> BaseEstimator:
        if len(df) == 0:
            raise Exception(f"Training data for model {model_name} is empty")

        model_hash = self._get_hash(model_type)

        cached_model = self._load_model_from_cache(model_name, model_hash)
        if cached_model:
            return cached_model

        model = self._get_estimator()
        grid_search_params = self._get_grid_search_params()

        if len(df) < self._config.k_fold_cv_splits:
            cv = 2
        else:
            cv = self._config.k_fold_cv_splits

        grid_search = GridSearchCV(
            estimator=model,
            param_grid=grid_search_params,
            scoring=self._get_scorer(),
            cv=cv,
            n_jobs=self._config.num_training_job_threads,
        )

        # we don't create a train/test split, because we want to use all data for training
        # and we don't care about overfitting, because we only want to predict execution time within the same domain
        X, y = df[feature_cols], df[target_col]

        grid_search.fit(X, y)
        score = grid_search.score(X, y)

        logger.info(
            f"Trained model {model_name} and found best parameters: {grid_search.best_params_} "
            f"with mean absolute percentage error (MEAP) {-score}%"
        )

        self._store_model_in_cache(model_name, model_hash, grid_search.best_estimator_)

        self._store_training_prediction_data(
            model_name=model_name,
            model_hash=model_hash,
            df=df,
            feature_cols=feature_cols,
            target_col=target_col,
            model=grid_search.best_estimator_,
        )
        return grid_search.best_estimator_

    def _store_model_prediction_cache(
        self, model_name: str, model_hash: str, predictions: Dict[Tuple, float]
    ) -> None:
        with InterProcessReaderWriterLock(
            f"{self._cache_dir}/{model_hash}_prediction_lock.file"
        ).write_lock():
            cache_file = (
                f"{self._cache_dir}/predictions/{model_hash}/{model_name}.csv.xz"
            )
            _store_prediction_file(cache_file, predictions)

    def _load_model_prediction_cache(
        self, model_name: str, model_hash: str
    ) -> Dict[Tuple, float]:
        with InterProcessReaderWriterLock(
            f"{self._cache_dir}/{model_hash}_prediction_lock.file"
        ).read_lock():
            if self._config.no_cache:
                return {}
            cache_file = (
                f"{self._cache_dir}/predictions/{model_hash}/{model_name}.csv.xz"
            )

            if not os.path.exists(cache_file):
                return {}

            logger.debug(f"Found model {model_name} predictions in cache")

            return _load_prediction_file(cache_file)

    def _get_model_prediction(
        self,
        model_name: str,
        model_type: ModelType,
        model: BaseEstimator,
        X: pd.DataFrame,
    ) -> Dict[Tuple, float]:
        X = X.copy()

        model_hash = self._get_hash(model_type)

        cached_predictions = self._load_model_prediction_cache(model_name, model_hash)
        if cached_predictions:
            return cached_predictions

        logger.info(f"Predicting execution time for model {model_name}")

        predictions_array = model.predict(X)

        # turn this into a dict, so we can store use it as a cache
        # the key is tuple for each row of X
        predictions = dict(zip([tuple(x) for x in X.values], predictions_array))

        self._store_model_prediction_cache(model_name, model_hash, predictions)

        return predictions

    def _train_general_models(
        self,
        compute_input_file: str,
        attention_input_file: str,
        all_reduce_input_file: str,
        send_recv_input_file: str,
    ) -> Dict[str, BaseEstimator]:

        compute_df = self._load_compute_df(compute_input_file)
        compute_df = self._get_compute_df_with_derived_features(compute_df)

        models = {}
        model_names = [
            "attn_pre_proj",
            "attn_post_proj",
            "mlp_up_proj",
            "mlp_down_proj",
            "mlp_act",
            "input_layernorm",
            "post_attention_layernorm",
            "attn_rope",
            "add",
        ]

        for model_name in model_names:
            logger.debug(
                f"Training model {model_name}, size of training data: {len(compute_df)}"
            )
            models[model_name] = self._train_model(
                model_name=model_name,
                model_type=ModelType.COMPUTE,
                df=compute_df,
                feature_cols=["num_tokens"],
                target_col=f"time_stats.{model_name}.median",
            )

        attention_df = self._load_attention_df(attention_input_file)
        attention_df = self._get_attention_df_with_derived_features(attention_df)

        model_names = [
            "attn_kv_cache_save",
        ]

        for model_name in model_names:
            models[model_name] = self._train_model(
                model_name=model_name,
                model_type=ModelType.COMPUTE,
                df=attention_df,
                feature_cols=["num_tokens"],
                target_col=f"time_stats.{model_name}.median",
            )

        if self._replica_config.num_pipeline_stages > 1:
            send_recv_df = self._load_send_recv_df(send_recv_input_file)
            send_recv_df = self._get_send_recv_df_with_derived_features(send_recv_df)

            models["send_recv"] = self._train_model(
                model_name="send_recv",
                model_type=ModelType.NETWORK,
                df=send_recv_df,
                feature_cols=["num_tokens"],
                target_col="time_stats.send_recv.median",
            )

        if self._replica_config.tensor_parallel_size > 1:
            all_reduce_df = self._load_all_reduce_df(all_reduce_input_file)
            all_reduce_df = self._get_all_reduce_df_with_derived_features(all_reduce_df)

            models["all_reduce"] = self._train_model(
                model_name="all_reduce",
                model_type=ModelType.NETWORK,
                df=all_reduce_df,
                feature_cols=["num_tokens"],
                target_col="time_stats.all_reduce.median",
            )

        if (
            self._replica_config.kv_parallel_size > 1
            and not self._config.disable_kvp_communication
        ):
            all_reduce_kvp_df = self._load_all_reduce_kvp_df(all_reduce_input_file)
            all_reduce_kvp_df = self._get_all_reduce_df_with_derived_features(
                all_reduce_kvp_df
            )

            models["all_reduce_kvp"] = self._train_model(
                model_name="all_reduce_kvp",
                model_type=ModelType.NETWORK,
                df=all_reduce_kvp_df,
                feature_cols=["num_tokens", "num_workers"],
                target_col="time_stats.all_reduce.median",
            )

        return models

    def _train_attention_layer_models(
        self, attention_input_file: str
    ) -> Dict[str, BaseEstimator]:
        attention_df = self._load_attention_df(attention_input_file)
        attention_df = self._get_attention_df_with_derived_features(attention_df)
        prefill_df = attention_df[~attention_df["is_decode"]]
        decode_df = attention_df[attention_df["is_decode"]]

        models = {}

        chunked_prefill_df = prefill_df[prefill_df["kv_cache_size"] > 0].copy()
        chunked_prefill_df["total_prefill_tokens"] = (
            chunked_prefill_df["kv_cache_size"]
            + chunked_prefill_df["prefill_chunk_size"]
        )

        models["attn_prefill"] = self._train_model(
            model_name="attn_prefill",
            model_type=ModelType.COMPUTE,
            df=prefill_df,
            feature_cols=["kv_cache_size", "prefill_chunk_size"],
            target_col="time_stats.attn.median",
        )

        models["attn_decode"] = self._train_model(
            model_name="attn_decode",
            model_type=ModelType.COMPUTE,
            df=decode_df,
            feature_cols=["batch_size", "kv_cache_size"],
            target_col="time_stats.attn.median",
        )

        return models

    def _train_models(self) -> Dict[str, BaseEstimator]:
        (
            compute_input_file,
            attention_input_file,
            all_reduce_input_file,
            send_recv_input_file,
        ) = self._get_input_files()

        models = self._train_general_models(
            compute_input_file,
            attention_input_file,
            all_reduce_input_file,
            send_recv_input_file,
        )
        models.update(self._train_attention_layer_models(attention_input_file))

        return models

    def _predict_for_general_models(self, models: Dict[str, Any]) -> Dict[str, Any]:
        predictions = {}

        model_names = [
            "attn_pre_proj",
            "attn_post_proj",
            "mlp_up_proj",
            "mlp_down_proj",
            "mlp_act",
            "attn_rope",
            "attn_kv_cache_save",
            "input_layernorm",
            "post_attention_layernorm",
            "add",
        ]

        num_token_range = np.arange(1, self._max_tokens + 1)
        X = pd.DataFrame({"num_tokens": num_token_range})

        for model_name in model_names:
            model = models[model_name]
            predictions[model_name] = self._get_model_prediction(
                model_name, ModelType.COMPUTE, model, X
            )

        network_model_names = []
        if self._replica_config.num_pipeline_stages > 1:
            network_model_names.append("send_recv")

        if self._replica_config.tensor_parallel_size > 1:
            network_model_names.append("all_reduce")

        for model_name in network_model_names:
            model = models[model_name]
            predictions[model_name] = self._get_model_prediction(
                model_name, ModelType.NETWORK, model, X
            )

        if self._replica_config.kv_parallel_size > 1:
            # we need to iterate over all possible number of workers and tokens
            num_workers_range = np.arange(1, self._replica_config.kv_parallel_size + 1)
            num_token_range = np.arange(1, self._max_tokens + 1)

            X = pd.DataFrame(
                list(product(num_token_range, num_workers_range)),
                columns=["num_tokens", "num_workers"],
            )

            if not self._config.disable_kvp_communication:
                model = models["all_reduce_kvp"]
                predictions["all_reduce_kvp"] = self._get_model_prediction(
                    "all_reduce_kvp", ModelType.NETWORK, model, X
                )

        return predictions

    def _predict_for_attention_layer_models(
        self, models: Dict[str, Any]
    ) -> Dict[str, Any]:
        predictions = {}

        decode_batch_size_range = np.arange(
            1, self._config.prediction_max_batch_size + 1
        )
        decode_kv_cache_size_range = np.arange(
            0,
            self._config.prediction_max_tokens_per_request + 1,
            self._config.kv_cache_prediction_granularity,
        )
        decode_prefill_chunk_size_range = [0]
        decode_batch_size, decode_kv_cache_size, decode_prefill_chunk_size = zip(
            *product(
                decode_batch_size_range,
                decode_kv_cache_size_range,
                decode_prefill_chunk_size_range,
            )
        )

        prefill_batch_size_range = [1]
        prefill_kv_cache_size_range = np.arange(
            0,
            self._config.prediction_max_tokens_per_request + 1,
            self._config.kv_cache_prediction_granularity,
        )
        prefill_prefill_chunk_size_range = np.arange(
            32, self._config.prediction_max_prefill_chunk_size + 1, 32
        )
        prefill_batch_size, prefill_kv_cache_size, prefill_prefill_chunk_size = zip(
            *product(
                prefill_batch_size_range,
                prefill_kv_cache_size_range,
                prefill_prefill_chunk_size_range,
            )
        )

        attention_df = pd.DataFrame(
            {
                "batch_size": decode_batch_size + prefill_batch_size,
                "kv_cache_size": decode_kv_cache_size + prefill_kv_cache_size,
                "prefill_chunk_size": decode_prefill_chunk_size
                + prefill_prefill_chunk_size,
            }
        )

        attention_df["is_decode"] = attention_df["prefill_chunk_size"] == 0
        attention_df["num_tokens"] = attention_df[
            ["prefill_chunk_size", "batch_size"]
        ].max(axis=1)
        attention_df["prefill_chunk_size_squared"] = (
            attention_df["prefill_chunk_size"] ** 2
        )

        prefill_df = attention_df[~attention_df["is_decode"]]
        decode_df = attention_df[attention_df["is_decode"]]
        chunked_prefill_df = prefill_df[prefill_df["kv_cache_size"] > 0].copy()
        chunked_prefill_df["total_prefill_tokens"] = (
            chunked_prefill_df["kv_cache_size"]
            + chunked_prefill_df["prefill_chunk_size"]
        )

        predictions["attn_prefill"] = self._get_model_prediction(
            "attn_prefill",
            ModelType.COMPUTE,
            models["attn_prefill"],
            prefill_df[["kv_cache_size", "prefill_chunk_size"]],
        )

        predictions["attn_decode"] = self._get_model_prediction(
            "attn_decode",
            ModelType.COMPUTE,
            models["attn_decode"],
            decode_df[["batch_size", "kv_cache_size"]],
        )

        return predictions

    def _predict_from_models(self, models: Dict[str, Any]) -> Dict[str, Any]:
        predictions = self._predict_for_general_models(models)
        predictions.update(self._predict_for_attention_layer_models(models))

        return predictions

    def _get_batch_decode_attention_params(self, batch: Batch) -> Tuple[int, int]:
        if hasattr(batch, "_decode_params"):
            return batch._decode_params

        decode_kv_cache_sizes = []

        for num_q_tokens, num_kv_tokens in zip(batch.num_q_tokens, batch.num_kv_tokens):
            if num_q_tokens != 1:
                continue
            decode_kv_cache_sizes.append(num_kv_tokens)

        if not decode_kv_cache_sizes:
            batch._decode_params = (0, 0)
            return batch._decode_params

        decode_batch_size = len(decode_kv_cache_sizes)
        decode_avg_kv_cache_size = int(np.mean(decode_kv_cache_sizes))
        decode_avg_kv_cache_size = (
            (
                decode_avg_kv_cache_size
                + self._config.kv_cache_prediction_granularity
                - 1
            )
            // self._config.kv_cache_prediction_granularity
        ) * self._config.kv_cache_prediction_granularity

        batch._decode_params = (decode_batch_size, decode_avg_kv_cache_size)

        return batch._decode_params

    def _get_batch_prefill_attention_params(
        self, batch: Batch
    ) -> List[Tuple[int, int]]:
        if hasattr(batch, "_prefill_params"):
            return batch._prefill_params

        prefill_params = []

        for num_q_tokens, num_kv_tokens in zip(batch.num_q_tokens, batch.num_kv_tokens):
            if num_q_tokens == 1:
                continue

            num_kv_tokens = (
                (num_kv_tokens + self._config.kv_cache_prediction_granularity - 1)
                // self._config.kv_cache_prediction_granularity
            ) * self._config.kv_cache_prediction_granularity

            prefill_params.append((num_kv_tokens, num_q_tokens))

        batch._prefill_params = prefill_params

        return prefill_params

    def _get_attention_layer_pre_proj_execution_time(self, batch: Batch) -> float:
        return self._predictions["attn_pre_proj"][(batch._total_num_q_tokens_rounded,)]

    def _get_attention_layer_post_proj_execution_time(self, batch: Batch) -> float:
        return self._predictions["attn_post_proj"][(batch._total_num_q_tokens_rounded,)]

    def _get_mlp_layer_up_proj_execution_time(self, batch: Batch) -> float:
        return self._predictions["mlp_up_proj"][(batch._total_num_q_tokens_rounded,)]

    def _get_mlp_layer_down_proj_execution_time(self, batch: Batch) -> float:
        return self._predictions["mlp_down_proj"][(batch._total_num_q_tokens_rounded,)]

    def _get_mlp_layer_act_execution_time(self, batch: Batch) -> float:
        return self._predictions["mlp_act"][(batch._total_num_q_tokens_rounded,)]

    def _get_attn_norm_layer_act_execution_time(self, batch: Batch) -> float:
        return self._predictions["input_layernorm"][
            (batch._total_num_q_tokens_rounded,)
        ]

    def _get_mlp_norm_layer_act_execution_time(self, batch: Batch) -> float:
        if not self._model_config.post_attn_norm:
            return 0

        return self._predictions["post_attention_layernorm"][
            (batch._total_num_q_tokens_rounded,)
        ]

    def _get_add_layer_act_execution_time(self, batch: Batch) -> float:
        return self._predictions["add"][(batch._total_num_q_tokens_rounded,)]

    def _get_tensor_parallel_communication_time(self, batch: Batch) -> float:
        return (
            self._predictions["all_reduce"][(batch._total_num_q_tokens_rounded,)]
            + self._config.nccl_cpu_launch_overhead_ms
            + self._config.nccl_cpu_skew_overhead_per_device_ms
            * self._replica_config.tensor_parallel_size**1.25
        )

    def _get_pipeline_parallel_communication_time(self, batch: Batch) -> float:
        try:
            return self._predictions["send_recv"][(batch._total_num_q_tokens_rounded,)]
        except KeyError as e:
            logger.error(f"Failed to get send_recv prediction for batch {batch}")
            raise e

    def _get_attention_rope_execution_time(self, batch: Batch) -> float:
        return self._predictions["attn_rope"][(batch._total_num_q_tokens_rounded,)]

    def _get_attention_kv_cache_save_execution_time(self, batch: Batch) -> float:
        # don't use round up to the nearest multiple of 8 here, because we want to
        # predict the execution time for the exact number of tokens
        return self._predictions["attn_kv_cache_save"][(batch._total_num_q_tokens,)]

    def _get_attention_decode_execution_time(self, batch: Batch) -> float:
        (
            decode_batch_size,
            decode_avg_kv_cache_size,
        ) = self._get_batch_decode_attention_params(batch)
        if decode_batch_size == 0:
            return 0

        return self._predictions["attn_decode"][
            (decode_batch_size, decode_avg_kv_cache_size)
        ] * (
            1
            + self._attention_decode_batching_overhead_fraction
            * int(decode_batch_size > 1)
        )

    def _get_attention_prefill_execution_time(self, batch: Batch) -> float:
        prefill_params = self._get_batch_prefill_attention_params(batch)

        if len(prefill_params) == 0:
            return 0

        total_time = 0

        for kv_cache_size, prefill_chunk_size in prefill_params:
            prefill_chunk_size = ((prefill_chunk_size + 31) // 32) * 32
            total_time += self._predictions["attn_prefill"][
                (kv_cache_size, prefill_chunk_size)
            ]

        return total_time

    @abstractmethod
    def _get_kv_parallel_communication_time(self, batch: Batch) -> float:
        if not self._config.disable_kvp_communication:
            return 0

        total_comm_time = 0

        for num_q_tokens, num_groups in zip(
            batch.num_q_tokens, batch.num_active_kvp_groups
        ):
            if num_q_tokens == 0:
                continue

            # round up to the nearest multiple of 8
            num_q_tokens = (num_q_tokens + 7) // 8 * 8

            total_comm_time += (
                self._predictions["all_reduce_kvp"][(num_q_tokens, num_groups)]
                + self._config.nccl_cpu_launch_overhead_ms
                + self._config.nccl_cpu_skew_overhead_per_device_ms * num_groups**1.25
            )

        return total_comm_time

    def _get_compute_model_config(self) -> Dict[str, Any]:
        return {
            "model_provider": str(self._config.get_type()),
            "profiling_datasource": self._config.model_profiling_dataset,
            "num_tensor_parallel_workers": self._replica_config.tensor_parallel_size,
            "k_fold_cv_splits": self._config.k_fold_cv_splits,
            "block_size": self._block_size,
            "max_tokens": self._max_tokens,
            "prediction_max_prefill_chunk_size": self._config.prediction_max_prefill_chunk_size,
            "max_batch_size": self._config.prediction_max_batch_size,
        }

    def _get_network_model_config(self) -> Dict[str, Any]:
        return {
            "model_provider": str(self._config.get_type()),
            "profiling_datasource": self._config.network_profiling_dataset,
            "tensor_parallel_size": self._replica_config.tensor_parallel_size,
            "is_multi_node": self._is_multi_node,
            "max_tokens": self._max_tokens,
            "disable_kvp_communication": self._config.disable_kvp_communication,
        }

    def to_dict(self) -> dict:
        return {
            "model_provider": str(self._config.get_type()),
            "model_name": self._replica_config.model_name,
            "model_profiling_dataset": self._config.model_profiling_dataset,
            "network_profiling_dataset": self._config.network_profiling_dataset,
            "tensor_parallel_size": self._replica_config.tensor_parallel_size,
            "is_multi_node": self._is_multi_node,
            "k_fold_cv_splits": self._config.k_fold_cv_splits,
            "max_tokens": self._max_tokens,
            "block_size": self._block_size,
            "prediction_max_prefill_chunk_size": self._config.prediction_max_prefill_chunk_size,
            "max_batch_size": self._config.prediction_max_batch_size,
        }



from vidur.config import ReplicaConfig
from vidur.entities.replica import Replica
from vidur.utils.param_counter import ParamCounter


class MemoryPlanner:
    def __init__(self, replica_config: ReplicaConfig, replica: Replica) -> None:
        self._param_counter = ParamCounter(replica_config)
        self._replica = replica

    def _get_kv_cache_memory_per_layer_per_request(self) -> int:
        return (
            2  # 2 bytes per float
            * 2  # one for key, one for value
            * self._replica.attention_head_dim
            * self._replica.kv_heads_per_tensor_parallel_worker
            * self._replica.max_request_tokens
        )

    def _get_parameter_memory_per_device(self) -> int:
        return 2 * self._param_counter.get_num_parameters_per_device()

    def _get_kv_cache_memory_per_device_per_request(self) -> int:
        return (
            self._get_kv_cache_memory_per_layer_per_request() * self._replica.num_layers
        )

    def get_max_batch_size(self) -> int:
        available_memory = (
            self._replica.total_memory_gb
            * 1024**3
            * (1 - self._replica.memory_margin_fraction)
        )
        parameter_memory_per_device = self._get_parameter_memory_per_device()
        kv_cache_memory_per_device_per_request = (
            self._get_kv_cache_memory_per_device_per_request()
        )

        memory_for_kv_cache = available_memory - parameter_memory_per_device
        number_of_requests = (
            memory_for_kv_cache // kv_cache_memory_per_device_per_request
        )

        assert (
            number_of_requests > 0
        ), "Not enough memory to store even a single request"

        return number_of_requests

    def get_max_request_slots(self) -> int:
        return self.get_max_batch_size() * self._replica.num_pipeline_stages



from typing import Dict, List

from vidur.entities.base_entity import BaseEntity
from vidur.entities.batch import Batch, check_completed, check_scheduled
from vidur.entities.batch_stage import BatchStage
from vidur.entities.execution_time import ExecutionTime
from vidur.entities.kv_parallel_batch_stage import KVParallelBatchStage
from vidur.entities.request import Request
from vidur.logger import init_logger

logger = init_logger(__name__)


class KVParallelBatch(BaseEntity):
    def __init__(
        self,
        replica_id: int,
        batch_mapping: Dict[int, Batch],  # map from kvp_group_id to batch
    ) -> None:
        self._id = KVParallelBatch.generate_id()
        self._replica_id = replica_id
        self._batch_mapping = batch_mapping

        self._scheduled_at = None
        self._completed_at = None
        self._scheduled = False
        self._completed = False

        self._requests = list(
            set(sum([x.requests for x in self._batch_mapping.values()], []))
        )

        self._req_num_q_token_map = {}
        for batch in self._batch_mapping.values():
            for request, num_q_tokens in zip(batch.requests, batch.num_q_tokens):
                if request.id in self._req_num_q_token_map:
                    assert self._req_num_q_token_map[request.id] == num_q_tokens, (
                        f"Request {request.id} has different number of q tokens in different kvp groups, "
                        f"{self._req_num_q_token_map[request.id]} and {num_q_tokens}, {self.to_dict()}"
                    )
                self._req_num_q_token_map[request.id] = num_q_tokens

    @property
    def requests(self) -> List[Request]:
        return self._requests

    @property
    def replica_id(self) -> int:
        return self._replica_id

    @property
    def batch_mapping(self) -> Dict[int, Batch]:
        return self._batch_mapping

    @property
    def creation_time(self) -> float:
        return self._creation_time

    @property
    @check_scheduled
    def scheduled_at(self) -> float:
        return self._scheduled_at

    @property
    @check_completed
    def completed_at(self) -> float:
        return self._completed_at

    @property
    def completed(self) -> bool:
        return self._completed

    @property
    def scheduled(self) -> bool:
        return self._scheduled

    @property
    def size(self) -> int:
        return sum(x.size for x in self._batch_mapping.values())

    @property
    def request_ids(self) -> List[int]:
        return sum([x.request_ids for x in self._batch_mapping.values()], [])

    @property
    def have_all_requests_completed(self) -> bool:
        return all(
            [x.have_all_requests_completed for x in self._batch_mapping.values()]
        )

    @property
    def preempted_requests(self) -> List[Request]:
        return sum([x.preempted_requests for x in self._batch_mapping.values()], [])

    @property
    def completed_requests(self) -> List[Request]:
        return sum([x.completed_requests for x in self._batch_mapping.values()], [])

    def get_num_q_tokens_for_request(self, request_id: int) -> int:
        return self._req_num_q_token_map[request_id]

    def on_schedule(
        self,
        time: float,
    ) -> None:
        self._scheduled_at = time
        self._scheduled = True

        for batch in self._batch_mapping.values():
            batch.on_schedule(time)

    def on_batch_end(self, time: float):
        self._completed = True
        self._completed_at = time

        for batch in self._batch_mapping.values():
            batch.on_batch_end(time)

    def to_batch_stage(
        self, stage_id: int, execution_time: ExecutionTime
    ) -> KVParallelBatchStage:
        batch_stage_mapping: Dict[int, BatchStage] = {
            kvp_group_id: batch.to_batch_stage(stage_id, execution_time)
            for kvp_group_id, batch in self._batch_mapping.items()
        }
        return KVParallelBatchStage(
            self._replica_id,
            batch_stage_mapping,
            stage_id,
            execution_time,
        )

    def to_dict(self) -> dict:
        return {
            "id": self._id,
            "size": self.size,
            "replica_id": self._replica_id,
            "scheduled_at": self._scheduled_at,
            "completed_at": self._completed_at,
            "scheduled": self._scheduled,
            "batches": [batch.to_dict() for batch in self._batch_mapping.values()],
        }



"""File to store names for different metrics captured"""

import enum


class OperationMetrics(enum.Enum):
    MLP_UP_PROJ = "mlp_up_proj"
    MLP_ACTIVATION = "mlp_activation"
    MLP_DOWN_PROJ = "mlp_down_proj"
    MLP_DOWN_PROJ_ALL_REDUCE = "mlp_down_proj_all_reduce"
    ATTN_PRE_PROJ = "attn_pre_proj"
    ATTN_POST_PROJ = "attn_post_proj"
    ATTN_POST_PROJ_ALL_REDUCE = "attn_post_proj_all_reduce"
    ATTN_PREFILL = "attn_prefill"
    ATTN_KV_CACHE_SAVE = "attn_kv_cache_save"
    ATTN_DECODE = "attn_decode"
    ATTN_ROPE = "attn_rope"
    PIPELINE_SEND_RECV = "pipeline_send_recv"
    ADD = "add"
    INPUT_LAYERNORM = "input_layernorm"
    POST_ATTENTION_LAYERNORM = "post_attention_layernorm"


class RequestMetricsTimeDistributions(enum.Enum):
    REQUEST_E2E_TIME = "request_e2e_time"
    REQUEST_E2E_TIME_NORMALIZED = "request_e2e_time_normalized"
    REQUEST_EXECUTION_TIME = "request_execution_time"
    REQUEST_EXECUTION_TIME_NORMALIZED = "request_execution_time_normalized"
    REQUEST_MODEL_EXECUTION_TIME = "request_model_execution_time"
    REQUEST_MODEL_EXECUTION_TIME_NORMALIZED = "request_model_execution_time_normalized"
    REQUEST_PREEMPTION_TIME = "request_preemption_time"
    REQUEST_SCHEDULING_DELAY = "request_scheduling_delay"
    REQUEST_EXECUTION_PLUS_PREEMPTION_TIME = "request_execution_plus_preemption_time"
    REQUEST_EXECUTION_PLUS_PREEMPTION_TIME_NORMALIZED = (
        "request_execution_plus_preemption_time_normalized"
    )
    PREFILL_TIME_E2E = "prefill_e2e_time"
    PREFILL_TIME_EXECUTION_PLUS_PREEMPTION = "prefill_time_execution_plus_preemption"
    PREFILL_TIME_EXECUTION_PLUS_PREEMPTION_NORMALIZED = (
        "prefill_time_execution_plus_preemption_normalized"
    )
    DECODE_TIME_EXECUTION_PLUS_PREEMPTION_NORMALIZED = (
        "decode_time_execution_plus_preemption_normalized"
    )


class TokenMetricsTimeDistribution(enum.Enum):
    DECODE_TOKEN_EXECUTION_PLUS_PREMPTION_TIME = (
        "decode_token_execution_plus_preemption_time"
    )


class RequestMetricsHistogram(enum.Enum):
    REQUEST_INTER_ARRIVAL_DELAY = "request_inter_arrival_delay"
    REQUEST_NUM_TOKENS = "request_num_tokens"
    REQUEST_PREFILL_TOKENS = "request_num_prefill_tokens"
    REQUEST_DECODE_TOKENS = "request_num_decode_tokens"
    REQUEST_PD_RATIO = "request_pd_ratio"
    REQUEST_NUM_RESTARTS = "request_num_restarts"


class BatchMetricsCountDistribution(enum.Enum):
    BATCH_NUM_TOKENS = "batch_num_tokens"
    BATCH_NUM_PREFILL_TOKENS = "batch_num_prefill_tokens"
    BATCH_NUM_DECODE_TOKENS = "batch_num_decode_tokens"
    BATCH_SIZE = "batch_size"


class BatchMetricsTimeDistribution(enum.Enum):
    BATCH_EXECUTION_TIME = "batch_execution_time"


class RequestCompletionMetricsTimeSeries(enum.Enum):
    REQUEST_ARRIVAL = "request_arrival"
    REQUEST_COMPLETION = "request_completion"


class TokenCompletionMetricsTimeSeries(enum.Enum):
    PREFILL_COMPLETIONS = "prefill_completion"
    DECODE_COMPLETIONS = "decode_completion"



import logging
from typing import List

import pandas as pd

from vidur.config import TraceRequestGeneratorConfig
from vidur.entities import Request
from vidur.request_generator.base_request_generator import BaseRequestGenerator

logger = logging.getLogger(__name__)


class TraceReplayRequestGenerator(BaseRequestGenerator):
    """
    Reads a trace csv file containing request arrival time, its prompt and completion token values to generate
    inter-request times, number of tokens.
    """

    def __init__(self, config: TraceRequestGeneratorConfig):
        super().__init__(config)

        # load into a pd dataframe
        self.trace_df = pd.read_csv(config.trace_file)
        # restrict trace_df to be a subset of rows that have the same date
        self.trace_df = self.trace_df[self.trace_df["Date"] == config.date]

        # scale prefill and decode tokens
        self.trace_df["PromptTokenCount"] = (
            self.trace_df["PromptTokenCount"] * config.prefill_scale_factor
        )
        self.trace_df["CompletionTokenCount"] = (
            self.trace_df["CompletionTokenCount"] * config.decode_scale_factor
        )

        # make sure all the prefill and decode counts are integers
        self.trace_df["PromptTokenCount"] = self.trace_df["PromptTokenCount"].astype(
            int
        )
        self.trace_df["CompletionTokenCount"] = self.trace_df[
            "CompletionTokenCount"
        ].astype(int)

        # make sure that there is at least one prefill and decode token
        self.trace_df["PromptTokenCount"] = self.trace_df["PromptTokenCount"].clip(
            lower=1
        )
        self.trace_df["CompletionTokenCount"] = self.trace_df[
            "CompletionTokenCount"
        ].clip(lower=1)

        # make sure the total does not exceed the max tokens, adjust the prefill tokens if needed
        total_tokens = (
            self.trace_df["PromptTokenCount"] + self.trace_df["CompletionTokenCount"]
        )
        diff_tokens = total_tokens - config.max_tokens
        diff_tokens = diff_tokens.clip(lower=0)
        self.trace_df["PromptTokenCount"] = (
            self.trace_df["PromptTokenCount"] - diff_tokens
        )

        assert all(
            self.trace_df["PromptTokenCount"] + self.trace_df["CompletionTokenCount"]
            <= config.max_tokens
        )

        # rescale the time to change QPS
        self.trace_df["Time"] = self.trace_df["Time"] * config.time_scale_factor

        # compute pd ratio and log the 25, 50, 75, 90, 95, 99 percentiles
        pd_ratio = (
            self.trace_df["PromptTokenCount"] / self.trace_df["CompletionTokenCount"]
        )
        logger.info(
            f"Loaded trace file {config.trace_file} with {len(self.trace_df)} requests"
        )
        logger.info(
            f"Prompt/decode token ratio stats\n:{pd_ratio.describe(percentiles=[0.25, 0.5, 0.75, 0.9, 0.95, 0.99])}"
        )

    def generate_requests(self) -> List[Request]:
        requests = []

        for _, row in self.trace_df.iterrows():
            request = Request(
                arrived_at=row["Time"],
                num_prefill_tokens=row["PromptTokenCount"],
                num_decode_tokens=row["CompletionTokenCount"],
            )

            requests.append(request)

        return requests



from abc import ABC, abstractmethod

from vidur.config import BaseRequestIntervalGeneratorConfig


class BaseRequestIntervalGenerator(ABC):

    def __init__(self, config: BaseRequestIntervalGeneratorConfig):
        self.config = config

    @abstractmethod
    def get_next_inter_request_time(self) -> float:
        pass



from vidur.types.base_int_enum import BaseIntEnum


class NodeSKUType(BaseIntEnum):
    A40_PAIRWISE_NVLINK = 1
    A100_PAIRWISE_NVLINK = 2
    H100_PAIRWISE_NVLINK = 3
    A100_DGX = 4
    H100_DGX = 5



from typing import List

from vidur.config import SyntheticRequestGeneratorConfig
from vidur.entities import Request
from vidur.request_generator.base_request_generator import BaseRequestGenerator
from vidur.request_generator.request_interval_generator_registry import (
    RequestIntervalGeneratorRegistry,
)
from vidur.request_generator.request_length_generator_registry import (
    RequestLengthGeneratorRegistry,
)
from vidur.types import RequestIntervalGeneratorType
from vidur.utils.random import set_seeds


class SyntheticRequestGenerator(BaseRequestGenerator):

    def __init__(self, config: SyntheticRequestGeneratorConfig):
        super().__init__(config)

        self.request_length_generator = RequestLengthGeneratorRegistry.get(
            self.config.length_generator_config.get_type(),
            self.config.length_generator_config,
        )
        self.request_interval_generator = RequestIntervalGeneratorRegistry.get(
            self.config.interval_generator_config.get_type(),
            self.config.interval_generator_config,
        )

    def _generate_next_request(self, last_arrived_at: float) -> Request:
        inter_request_time = (
            self.request_interval_generator.get_next_inter_request_time()
        )
        if inter_request_time is None:
            return None
        arrived_at = last_arrived_at + inter_request_time

        (
            prefill_tokens,
            decode_tokens,
        ) = self.request_length_generator.get_next_num_tokens()

        if prefill_tokens is None or decode_tokens is None:
            return None

        return Request(
            arrived_at=arrived_at,
            num_prefill_tokens=int(prefill_tokens),
            num_decode_tokens=int(decode_tokens),
        )

    def _generate_requests(self) -> List[Request]:
        requests = []

        current_time = 0

        # first priority is duration
        if self.config.duration is not None:
            while current_time < self.config.duration:
                request = self._generate_next_request(current_time)
                current_time = request.arrived_at
                requests.append(request)
        elif self.config.num_requests is not None:
            for _ in range(self.config.num_requests):
                request = self._generate_next_request(current_time)
                current_time = request.arrived_at
                requests.append(request)
        else:
            assert (
                self.config.interval_generator_config.get_type()
                == RequestIntervalGeneratorType.TRACE
            )

            while True:
                request = self._generate_next_request(current_time)
                if request is None:
                    break
                current_time = request.arrived_at
                requests.append(request)

        return requests

    def generate_requests(self) -> List[Request]:
        assert (
            self.config.num_requests
            or self.config.duration
            or self.config.interval_generator_config.get_type()
            == RequestIntervalGeneratorType.TRACE
        )

        set_seeds(self.config.seed)

        requests = self._generate_requests()

        # sort requests by arrival time
        requests.sort(key=lambda x: x.arrived_at)
        # remove any requests that arrived after the time limit
        if self.config.duration is not None:
            requests = [
                request
                for request in requests
                if request.arrived_at < self.config.duration
            ]

        return requests



from abc import ABC, abstractmethod
from typing import Tuple

from vidur.config import BaseRequestLengthGeneratorConfig


class BaseRequestLengthGenerator(ABC):

    def __init__(self, config: BaseRequestLengthGeneratorConfig):
        self.config = config

    @abstractmethod
    def get_next_num_tokens(self) -> Tuple[float, float]:
        pass



from itertools import product

import pandas as pd
import streamlit as st

from vidur.config_optimizer.analyzer.constants import AXIS_COLS
from vidur.config_optimizer.analyzer.dashboard.utils import add_small_divider, plot_cdf


def render_comparison_points(df):
    configs_to_remove = []
    rows_to_render = []
    COMPARE_COLS = ["Model", "Trace"] + list(AXIS_COLS.values())
    for idx, point in enumerate(st.session_state.comparison_points):
        config_row = df[(df[COMPARE_COLS] == pd.Series(point)).all(axis=1)]
        if config_row.empty:
            st.warning(f"Configuration {point} not found in the dataset")
            configs_to_remove.append(point)
            continue

        rows_to_render.append((idx, point, config_row))

    for point in configs_to_remove:
        st.session_state.comparison_points.remove(point)

    if not rows_to_render:
        st.warning("No configurations to compare")
        return

    cols = st.columns([1] * len(rows_to_render))
    num_cols = len(rows_to_render)

    for idx, (idx, point, config_row) in enumerate(rows_to_render):
        col = cols[idx]
        with col:
            container = st.container(border=True)
            with container:
                title_col, cancel_col, _ = st.columns(
                    [100 - 5 * num_cols, 5, 5 * num_cols]
                )
                with title_col:
                    st.markdown(f"### {config_row['Name'].values[0]}")
                with cancel_col:
                    remove_button = st.button("✖️", key=f"remove_{idx}")

                add_small_divider()

                st.markdown("##### Configuration Details")
                for axis_col, value in point.items():
                    st.write(f"**{axis_col}**: {value}")

                add_small_divider()

                st.markdown("##### Performance Metrics")

                st.write(f"**TTFT (P95)**: {config_row['TTFT_95%'].values[0]:.2f} s")
                st.write(f"**TBT (P99)**: {config_row['TBT_99%'].values[0]:.2f} ms")
                st.write(
                    f"**QPS per Dollar**: {config_row['QPS per Dollar'].values[0]:.2f}"
                )
                st.write(f"**MFU**: {config_row['mfu_mean'].values[0]:.2f}%")
                st.write(
                    f"**Memory Usage**: {config_row['memory_usage_mean'].values[0]:.2f}%"
                )
                st.write(
                    f"**Busy Time**: {config_row['busy_time_percent_mean'].values[0]:.2f}%"
                )

                add_small_divider()

            if remove_button:
                st.session_state.comparison_points.remove(point)
                st.experimental_rerun()

    add_small_divider()
    st.markdown("### CDFs")
    config_rows = [config_row for _, _, config_row in rows_to_render]
    config_df = pd.concat(config_rows)
    plot_cdf(config_df, y="TTFT_cdf", y_name="TTFT", color="Name")
    plot_cdf(config_df, y="TBT_cdf", y_name="TBT", color="Name")
    plot_cdf(config_df, y="batch_size_cdf", y_name="Batch Size", color="Name")
    plot_cdf(
        config_df, y="batch_num_tokens_cdf", y_name="Batch Num Tokens", color="Name"
    )


def render_config_comparison_page(df):
    st.markdown("## Config Comparison")
    st.markdown(
        'This tool enabled one-on-one performance comparison between different configurations. You can either manually add configuration on this page or use the **"Add to Configuration"** button on other pages to quickly populate the comparison.'
    )
    add_small_divider()
    st.markdown("### Select configurations to compare")

    COMPARISON_COLS = ["Model", "Trace"] + list(AXIS_COLS.values())

    comparison_features = {}
    with st.form(key="comparison_point_form"):
        col1, col2 = st.columns(2)

        for i, axis_col in enumerate(COMPARISON_COLS):
            col = col1 if i % 2 == 0 else col2
            with col:
                comparison_features[axis_col] = st.multiselect(
                    axis_col, df[axis_col].unique(), key=f"{axis_col}_selector"
                )

        # Submit button for the form
        submit_button = st.form_submit_button(label="Add Comparison Points")

    comparison_points = []
    # take the product of all the selected features
    for point in product(*comparison_features.values()):
        comparison_points.append(dict(zip(COMPARISON_COLS, point)))

    if submit_button:
        for comparison_point in comparison_points:
            if comparison_point not in st.session_state.comparison_points:
                st.session_state.comparison_points.append(comparison_point)
            else:
                st.warning(f"{comparison_point} is already added for comparison.")

    if st.session_state.comparison_points:
        title_col, clear_col = st.columns([0.9, 0.1])
        st.markdown("--- \n\n")
        with title_col:
            st.markdown("### Comparison Results")
        with clear_col:
            clear_button = st.button("Clear All", key="clear_all")
            if clear_button:
                st.session_state.comparison_points = []
                return

        render_comparison_points(df)
